<?php

namespace Drupal\pbs_media_manager;

use DateTime;
use DateTimeZone;
use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Language\Language;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\State\StateInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Drupal\media\Entity\Media;

class AssetManager extends ApiContentManagerBase {

  /**
   * Show manager.
   *
   * @var \Drupal\pbs_media_manager\ShowManager
   */
  protected ShowManager $showManager;

  /**
   * Season manager.
   *
   * @var \Drupal\pbs_media_manager\SeasonManager
   */
  protected SeasonManager $seasonManager;

  /**
   * Episode manager.
   *
   * @var \Drupal\pbs_media_manager\EpisodeManager
   */
  protected EpisodeManager $episodeManager;

  /**
   * Special manager.
   *
   * @var \Drupal\pbs_media_manager\SpecialManager
   */
  protected SpecialManager $specialManager;

  /**
   * @inheritDoc
   */
  public static function getEntityTypeId(): string {
    return 'media';
  }

  /**
   * @inheritDoc
   */
  public function getBundleId(): string {
    return $this->config->get('videos.video_media_type');
  }

  /**
   * @inheritDoc
   */
  public function getLastUpdateTime(): DateTime {
    return new DateTime();
  }

  /**
   * @inheritDoc
   */
  public function setLastUpdateTime(DateTime $time): void {
  }

  /**
   * @inheritDoc
   */
  public function resetLastUpdateTime(): void {
  }

  /**
   * @inheritDoc
   */
  public static function getAutoUpdateConfigName(): string {
    return '';
  }

  /**
   * @inheritDoc
   */
  public static function getAutoUpdateIntervalConfigName(): string {
    return '';
  }

  /**
   * {@inheritdoc}
   */
  public static function getQueueName(): string {
    return 'pbs_media_manager.queue.assets';
  }

  /**
   * ApiContentManagerBase constructor.
   *
   * {@inheritdoc}
   */
  public function __construct(
    ApiClient $client,
    EntityTypeManagerInterface $entity_type_manager,
    ModuleHandlerInterface $module_handler,
    ConfigFactoryInterface $config_factory,
    LoggerChannelInterface $logger,
    QueueFactory $queue_factory,
    StateInterface $state,
    FileSystemInterface $file_system,
    ShowManager $show_manager,
    SeasonManager $season_manager,
    EpisodeManager $episode_manager,
    SpecialManager $special_manager
  ) {
    parent::__construct($client, $entity_type_manager, $module_handler,
      $config_factory, $logger, $queue_factory, $state, $file_system);
    $this->showManager = $show_manager;
    $this->seasonManager = $season_manager;
    $this->episodeManager = $episode_manager;
    $this->specialManager = $special_manager;
  }

  /**
   * @inheritDoc
   * @throws \OpenPublicMedia\PbsMediaManager\Exception\BadRequestException|\Drupal\Core\Entity\EntityStorageException
   */
  public function updateQueue(DateTime $since = NULL): bool {
    $logging = $this->config->get('logging.operations');

    if (empty($since)) {
      $since = $this->getLastUpdateTime();
    }

    // Initialize counter for tracking items added to queue.
    $count = 0;

    // Check to see if we're subscribed to specific shows.
    $subscription = FALSE;
    $show_vocabulary = $this->config->get('shows.show_vocabulary');
    if (!empty($show_vocabulary) && $show_vocabulary != 'unused') {
      $subscription = TRUE;
    }

    // Check to see if we're using the changelog. If not, asset updates are
    // handled when updating parent items.
    $changelog = $this->config->get('shows.queue.changelog');
    if ($changelog) {
      // Get UTC datetime for "since" argument to changelog endpoint.
      $utc = new DateTimeZone('UTC');
      $since_utc = clone($since);
      $since_utc->setTimezone($utc);

      // Build the query to get the updated shows from the changelog.
      $query = [
        'type' => 'asset',
        'since' => $since_utc->format(self::API_DATETIME_FORMAT),
        'sort' => '-timestamp',
        'assets' => TRUE,
      ];


      // If we are subscribed to specific shows, we need to get those GUIDs.
      if ($subscription) {
        // Get the shows.
        $show_guids = $this->showManager->getSubscribedShows($show_vocabulary);
      }
      // Otherwise, we need to get all the shows from the database.
      else {
        // Get existing Show nodes keyed by Media Manager GUID.
        $shows = $this->showManager->getShowNodes();
        $shows = array_filter($shows, [$this, 'nodeHasGuid']);
        $shows = array_combine(array_map([$this, 'getNodeGuid'], $shows),
          $shows);

        // Flip it so we just have an array of guids.
        $show_guids = array_keys($shows);

      }

      $upsert_guids = [];
      $delete_guids = [];
      if (!empty($show_guids)) {
        foreach ($show_guids as $guid) {
          // Limit the changelog query to one show at a time. This way we're
          // only getting assets we have access to.
          $query['id'] = $guid;
          // Get the changelog.
          $changelog_result = $this->client->getChangelog($query);
          // Divide into upsert vs. delete.
          $changelog_result = $this->sortChangelog($changelog_result);
          $upsert_guids = array_merge($upsert_guids, $changelog_result['upsert']);
          $delete_guids = array_merge($delete_guids, $changelog_result['delete']);
        }
      }

      // Cycle through the created/updated objects and add them to the queue.
      foreach ($upsert_guids as $guid) {
        // Query the API for the full record for the asset.
        if ($returned_item = $this->client->getAsset($guid)) {
          // Add it to the queue.
          $this->getQueue()->createItem($returned_item);
          $count++;
        }
      }

      // Update the logger with the total number of assets added.
      if (!$logging['queue']) {
        $this->logger->notice('{count} total assets have been added to the queue
        .', ['count' => $count ]);
      }

      // Cycle through the deleted assets and delete the associated media
      // entity.
      foreach ($delete_guids as $guid) {
        if ($asset = $this->getAssetByGuid($guid)) {
          if (!$logging['changed']) {
            $this->logger->notice($this->t('The asset @title will be deleted', [
              '@title' => $asset->getName(),
            ]));
          }
          $asset->delete();
        };
      }
    }

    return TRUE;
  }

  /**
   *
   * Adds or updates Drupal video media data with Media Manager API data.
   *
   * @param object $item
   *   Media Manager API data for an Asset.
   * @param bool $force
   *   Whether to "force" the update. If FALSE (default) and a current
   *   Video Content node is found with a matching or newer "last_updated"
   *   value, the existing node will not be updated. If TRUE, "last_updated" is
   *   ignored.
   *
   * @return \Drupal\media\Entity\Media|null
   *   The Drupal video media entity.
   */
  public function addOrUpdateAssetContent(object $item, bool $force = FALSE): ?Media {
    // Check for required configuration.
    $video_media_type = $this->config->get('videos.video_media_type');
    if (empty($video_media_type)) {
      $this->logger->error('Please configure the video media type.');
      return NULL;
    }
    $mappings = $this->config->get('videos.mappings');

    $media_storage = $this->entityTypeManager->getStorage('media');
    $id_field = $mappings['id'];
    $title_field = $mappings['title'];
    $attributes = $item->attributes;
    if ($media_video = $media_storage->loadByProperties([
      $id_field => $item->id,
    ])) {
      if (count($media_video) > 1) {
        $this->logger->error(
          $this->t('More than one video with the @id (@title) exists. Please delete duplicate videos.', [
            '@id' => $item->id,
            '@title' => $attributes->title,
          ]));
        return NULL;
      }
      $media_video = reset($media_video);
    }
    // Otherwise, create download the file and create a new video media entity.
    else {
      $media_video = $media_storage->create([
        $title_field => $attributes->title,
        'bundle' => $video_media_type,
        'langcode' => Language::LANGCODE_NOT_SPECIFIED,
      ]);
      $media_video->enforceIsNew();
    }
    $operation = $this->getOperation($item, $media_video, $mappings, $force);
    $logging = $this->config->get('logging.operations');

    if ($operation == 'configured') {
      $this->logger->error($this->t('Please map the title, id, and updated_at fields for videos.'));
      return NULL;
    }

    if ($operation == 'skipped') {
      if (!$logging['skipped']) {
        $this->logger->debug($this->t('The asset @title has not changed', [
          '@title' => $media_video->getName(),
        ]));
      }
      return $media_video;
    }
    if ($operation == 'ignored') {
      if (!$logging['ignored']) {
        $this->logger->debug($this->t('The asset @id was ignored due to a hook_pbs_media_manager_operation_alter() implementation.', [
          '@id' => $item->id,
        ]));
      }
      return NULL;
    }

    // Media Manager video ID.
    $media_video->set($id_field, $item->id);

    // Media Manager updated date.
    $updated_dt = self::dateTimeNoMicroseconds($item->attributes->updated_at);
    $updated_at_field = $mappings['updated_at'];
    $media_video->set(
      $updated_at_field,
      $updated_dt->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT)
    );

    // Availability updated dates are father straightforward.
    $availabilities = $attributes->availabilities;
    $station_members_updated_at = $mappings['station_members_updated_at'];
    if (!empty($station_members_updated_at) && $station_members_updated_at !== 'unused') {
      $station_updated_dt = self::dateTimeNoMicroseconds($availabilities->station_members->updated_at);
      $media_video->set(
        $station_members_updated_at,
        $station_updated_dt->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT)
      );
    }
    $public_updated_at = $mappings['public_updated_at'];
    if (!empty($public_updated_at) && $public_updated_at !== 'unused') {
      $pub_updated_dt = self::dateTimeNoMicroseconds($availabilities->public->updated_at);
      $media_video->set(
        $public_updated_at,
        $pub_updated_dt->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT)
      );
    }
    $all_members_updated_at = $mappings['all_members_updated_at'];
    if (!empty($all_members_updated_at) && $all_members_updated_at !== 'unused') {
      $all_updated_dt = self::dateTimeNoMicroseconds($availabilities->all_members->updated_at);
      $media_video->set(
        $all_members_updated_at,
        $all_updated_dt->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT)
      );
    }

    $media_video = $this->addAvailabilityStartEnd(
      $media_video,
      $mappings['station_members_avail_start'],
      $mappings['station_members_avail_end'],
      $availabilities->station_members->start,
      $availabilities->station_members->end
    );
    $media_video = $this->addAvailabilityStartEnd(
      $media_video,
      $mappings['public_avail_start'],
      $mappings['public_avail_end'],
      $availabilities->public->start,
      $availabilities->public->end
    );
    $media_video = $this->addAvailabilityStartEnd(
      $media_video,
      $mappings['all_members_avail_start'],
      $mappings['all_members_avail_end'],
      $availabilities->all_members->start,
      $availabilities->all_members->end
    );

    // Process images.
    if ($images = $attributes->images) {
      foreach ($images as $image) {
        if (!empty($mappings[$image->profile]) && $image_field = $mappings[$image->profile]) {
          if ($image_field != 'unused') {
            // Check to see if this item has an image already.
            $media_image = NULL;
            if ($media_video->get($image_field)->target_id) {
              // Get the existing image entity.
              $image_id = $media_video->get($image_field)->target_id;
              $media_storage = $this->entityTypeManager->getStorage('media');
              $media_image = $media_storage->load($image_id);
            }
            if ($media_id = $this->addOrUpdateMediaImage(
              $image,
              $attributes->title . ": " . $image->profile,
              'video',
              $media_image,
              $force
            )) {
              $media_video->set($image_field, ['target_id' => $media_id]);
            }
          }
        }
      }
    }

    $remaining_fields = [
      'description_long',
      'description_short',
      'duration',
      'encored_on',
      'object_type',
      'premiered_on',
      'slug',
      'title_sortable',
      'player_code',
      'title',
      'language',
      'drm_enabled'
    ];
    foreach ($remaining_fields as $field) {
      $mapped_field = !empty($mappings[$field]) ? $mappings[$field] : NULL;
      if ($attributes->{$field} &&
        $mapped_field && $mapped_field !== 'unused' &&
        $media_video->hasField($mapped_field)
      ) {
        $media_video->set($mapped_field, $attributes->{$field});
      }
    }

    if ($uid = $this->config->get('api.queue_user')) {
      $media_video->setOwnerId($uid);
    }

    if (!$logging['changed']) {
      $this->logger->notice($this->t('The asset @title has been @operation', [
        '@title' => $media_video->getName(),
        '@operation' => $operation,
      ]));
    }

    // We need to save the video entity so the entity ID is available when
    // adding content references.
    $media_video->save();

    // Add the content references.
    $media_video = $this->addContentReferences($media_video, $attributes->parent_tree);

    // Resave.
    $media_video->save();

    return $media_video;
  }

  /**
   * Adds configured availability data to media video.
   *
   * @param \Drupal\media\Entity\Media $media_video
   *   The Drupal video media entity.
   * @param string $start_field
   *   The start field.
   * @param string $end_field
   *   The end field.
   * @param string|null $avail_start
   *   The availabilities start date string.
   * @param string|null $avail_end
   *   The availabilities end date string.
   *
   * @return \Drupal\media\Entity\Media
   *   The Drupal video media entity.
   */
  protected function addAvailabilityStartEnd(Media $media_video, string
  $start_field, string $end_field, ?string $avail_start, ?string $avail_end
  ): Media {

    // Begin with the availability start.
    if (!empty($start_field) && $start_field!=='unused'){
      if (!empty($avail_start)) {
        $start_dt = self::dateTimeNoMicroseconds($avail_start);
        // Next, check the availability end.
        if (!empty($end_field) && $end_field !== 'unused') {
          if (!empty($avail_end)){
            $end_dt = self::dateTimeNoMicroseconds($avail_end);
            // TODO: Remove this check once this issues is committed to core:
            // https://www.drupal.org/project/drupal/issues/2794481
            if ($this->moduleHandler->moduleExists('optional_end_date')) {
              // If start and end need to go in the same field add them together.
              $media_video->set(
                $start_field,
                [
                  'value' => $start_dt->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT),
                  'end_value' => $end_dt->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT),
                ]
              );
              return $media_video;
            }
            // Otherwise, add the end in its own field.
            else {
              $media_video->set(
                $end_field,
                $end_dt->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT)
              );
            }
          } else {
            $media_video->set(
              $end_field,
              NULL
            );
          }
        }
        $media_video->set(
          $start_field,
          $start_dt->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT)
        );
      } else {
        $media_video->set(
          $start_field,
          NULL
        );
      }
    }
    return $media_video;
  }

  /**
   * Adds show, season, and episode references to media video.
   *
   * TODO: Determine how to handle missing parents.
   *
   * @param \Drupal\media\Entity\Media $media_video
   *   The Drupal video media entity.
   * @param object $parent_tree
   *   Parent tree data from the Media Manager Asset.
   *
   * @return \Drupal\media\Entity\Media
   *   The Drupal video media entity.
   */
  protected function addContentReferences(Media $media_video, object $parent_tree): Media {
    $video_references = $this->config->get('videos.references');
    $tree = $this->parseParentTree($parent_tree);
    // Allows episode nodes to be referenced back from media videos
    // when the parent tree has a special but doesn't have an episode.
    if (!isset($tree['episode']) && isset($tree['special'])) {
      $tree['episode'] = $tree['special'];
    }
    // Cycle though the parent tree.
    foreach ($tree as $type => $guid) {
      $bundle = $this->config->get($type . 's.drupal_' . $type . '_content');
      if (isset($video_references[$type]) && $ref = $video_references[$type]) {
        if (!empty($guid) && !empty($bundle) && $ref != 'unused') {
          // Get the associated parent content entity.
          $node = $this->getNodeByGuid($guid, $bundle, $type . 's');
          // If no associated parent content entity exists, try to create it.
          if (empty($node)) {
            // Get functions.
           $functions = $this->parentFunctions();
           $manager = $functions[$type]['manager'];

            // Get the item from the API.
            $item = $this->{$manager}->{$functions[$type]['get']}($guid);

            // Add the item to the database.
            if (is_object($item)) {
              $this->{$manager}->{$functions[$type]['add']}($item, ['force' => TRUE]);
              // Get the resulting node.
              $node = $this->getNodeByGuid($guid, $bundle, $type . 's');
            }
          }

          // If we have a parent entity, add the references.
          if ($node && !empty($node->id())) {
            $media_video->set($ref, ['target_id' => $node->id()]);
            // Check to see if we're using the changelog. If so, we also
            // need to update the parent tree to reference back to this asset.
            $changelog = $this->config->get('shows.queue.changelog');
            if ($changelog) {
              // Get the video field.
              $video_field = $this->config->get($type . 's.mappings.video');
              // Get referenced assets from parent.
              $existing_video_references = $node->get($video_field)->referencedEntities();

              // Build an array of referenced asset IDs.
              $existing_video_targets = [];
              foreach ($existing_video_references as $existing_video_reference) {
                $existing_video_targets[$existing_video_reference->id] = ['target_id' => $existing_video_reference->id()];
              }
              // If the asset isn't already referenced, add it.
              if (!isset($existing_video_targets[$media_video->id()])) {
                $existing_video_targets[$media_video->id()] =
                  ['target_id' => $media_video->id()];
                // Re-index the array to reset the deltas.
                $existing_video_targets = array_values($existing_video_targets);
                // Set the field and save the parent node.
                $node->set($video_field, $existing_video_targets);
                $node->save();
              }
            }
          }
        }
      }
    }
    return $media_video;
  }

  /**
   * Helper to retrieve parent entity functions.
   *
   * @return array
   */
  private function parentFunctions() {
    return [
      'show' => [
        'manager' => 'showManager',
        'get' => 'getShow',
        'add' => 'addOrUpdateShow'
      ],
      'season' => [
        'manager' => 'seasonManager',
        'get' => 'getSeason',
        'add' => 'addOrUpdateSeason'
      ],
      'episode' => [
        'manager' => 'episodeManager',
        'get' => 'getEpisode',
        'add' => 'addOrUpdateEpisode'
      ],
      'special' => [
        'manager' => 'specialManager',
        'get' => 'getSpecial',
        'add' => 'addOrUpdateSpecial'
      ]
    ];
  }


  /**
   * Parse a Media Manager API Asset parent tree.
   *
   * @param object $parent_tree
   *   Parent tree data from the Media Manager Asset.
   *
   * @return array
   *   Ordered array keyed by parent object types and their Media Manager ID
   *   values. E.g. --
   *
   * @code
   * <?php
   * $tree = [
   *   'episode' => 'e358cacc-86e6-46a8-9269-dc72c8e34143',
   *   'season' => 'da41c844-f87e-4c89-ae40-f41c72ebdd0f',
   *   'show' => '294a8c40-1d69-4711-9163-5d17872b40e2',
   *   'franchise' => 'e08bf78d-e6a3-44b9-b356-8753d01c7327',
   * ];
   * @endcode
   */
  protected function parseParentTree(object $parent_tree): array {
    $tree = [];

    // The initial parent will only ever be "franchise", "show", "special",
    // "season", or "episode".
    if (isset($parent_tree->type)) {
      $tree[$parent_tree->type] = $parent_tree->id;
    }
    $branch = $parent_tree->attributes;

    // "episode" will have a "season".
    if (isset($branch->season)) {
      $tree['season'] = $branch->season->id;
      $branch = $branch->season->attributes;
    }

    // "special" or "season" will have a "show".
    if (isset($branch->show)) {
      $tree['show'] = $branch->show->id;
      $branch = $branch->show->attributes;
    }

    // "show" _might_ have a "franchise".
    if (isset($branch->franchise)) {
      $tree['franchise'] = $branch->franchise->id;
    }

    return $tree;
  }

}
