<?php

namespace Drupal\pbs_media_manager;

use DateTime;
use DateTimeZone;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Queue\QueueFactory;

class ChangelogManager {

  /**
   * PBS Media Manager API client wrapper.
   *
   * @var \Drupal\pbs_media_manager\ApiClient
   *
   * @see \OpenPublicMedia\PbsMediaManager\Client
   */
  protected $client;

  /**
   * Media Manager logger.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Queue service.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueFactory;

  /**
   * Show manager.
   *
   * @var \Drupal\pbs_media_manager\ShowManager
   */
  protected $showManager;

  /**
   * Season manager.
   *
   * @var \Drupal\pbs_media_manager\SeasonManager
   */
  protected $seasonManager;

  /**
   * Episode manager.
   *
   * @var \Drupal\pbs_media_manager\EpisodeManager
   */
  protected $episodeManager;

  /**
   * Special manager.
   *
   * @var \Drupal\pbs_media_manager\SpecialManager
   */
  protected $specialManager;


  /**
   * Asset manager.
   *
   * @var \Drupal\pbs_media_manager\AssetManager
   */
  protected $assetManager;

  /**
   * Datetime format for Media Manager datetime data.
   */
  const API_DATETIME_FORMAT = 'Y-m-d\TH:i:s.u\Z';

  /**
   * ApiContentManagerBase constructor.
   *
   * @param \Drupal\pbs_media_manager\ApiClient $client
   *   Media Manager API client service.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   *   Logger channel service.
   * @param \Drupal\Core\Queue\QueueFactory $queue_factory
   *   Queue factory service.
   * @param \Drupal\pbs_media_manager\ShowManager $show_manager
   *   PBS Media Manager show manager.
   * @param \Drupal\pbs_media_manager\SeasonManager $season_manager
   *   PBS Media Manager season manager.
   * @param \Drupal\pbs_media_manager\EpisodeManager $episode_manager
   *   PBS Media Manager episode manager.
   * @param \Drupal\pbs_media_manager\SpecialManager $special_manager
   *   PBS Media Manager special manager.
   * @param \Drupal\pbs_media_manager\AssetManager $asset_manager
   *   PBS Media Manager special manager.
   */
  public function __construct(
    ApiClient $client,
    LoggerChannelInterface $logger,
    QueueFactory $queue_factory,
    ShowManager $show_manager,
    SeasonManager $season_manager,
    EpisodeManager $episode_manager,
    SpecialManager $special_manager,
    AssetManager $asset_manager
  ) {
    $this->client = $client;
    $this->logger = $logger;
    $this->queueFactory = $queue_factory;
    $this->showManager = $show_manager;
    $this->seasonManager = $season_manager;
    $this->episodeManager = $episode_manager;
    $this->specialManager = $special_manager;
    $this->assetManager = $asset_manager;
  }


  /*
   * This is being called by cron at the appropriate update interval.
   *
   * Parameters passed include the types and the last update time.
   *
   * Process:
   * 1. Hit the changelog endpoint for each type passed, filtered by the last
   *  update time.
   * 2. If the action is create or update, use the appropriate Manager class
   * to get the object from the API, and if access is granted, add it to the
   * queue for that type.
   * 3. If the action is delete, add it to the new purge queue for later
   * deletion.
   */


  /**
   * @param \DateTime|NULL $since
   *   The last update time.
   * @param array|NULL $types
   *   Which resource types to retrieve from the API.
   *
   * @return bool
   * @throws \OpenPublicMedia\PbsMediaManager\Exception\BadRequestException
   */
  public function updateQueues(DateTime $since = NULL, array $types = NULL):
  bool {
    var_dump('updatequeues is running');
    $start_dt = new DateTime();

    if (empty($since)) {
      $since = $this->showManager->getLastUpdateTime();
      // Get UTC datetime for "since" argument to changelog endpoint.
      $utc = new DateTimeZone('UTC');
      $since->setTimezone($utc);
    }

    // Match the changelog type to the API call in the client.
    $api_call = [
      'show' => 'getShow',
      'season' => 'getSeason',
      'episode' => 'getEpisode',
      'special' => 'getSpecial',
    ];
    // Match the changelog type to the manager class.
    $manager = [
      'show' => 'showManager',
      'season' => 'seasonManager',
      'episode' => 'episodeManager',
      'special' => 'specialManager',
    ];

    foreach ($types as $type) {
      // Initialize arrays to hold results.
      $guids = ['upsert' => [], 'delete' => []];
      // Initialize array to hold show guids for future checking.
      if ($type == 'show') {
        $show_guids = [];
      }

      // Set up the changelog query to limit by resource type and to the last
      // time the update ran.
      $query = [
        'type' => $type,
        'since' => $since->format(self::API_DATETIME_FORMAT),
        'sort' => '-timestamp',
      ];

      // Get the changelog.
      $objects = $this->client->getChangelog($query);

      // Sort the returned objects into deleted or created/updated.
      foreach ($objects as $object) {
        // Add items marked for deletion to an array for future processing.
        if ($object->attributes->action === 'delete') {
          $guids['delete'][$object->id] = $object->id;
        }
        // All other items should be created or updated.
        else {
          $guids['upsert'][$object->id] = $object->id;
        }
      }

      // Cycle through the created/updated objects to determine if we have
      // access.
      foreach ($guids['upsert'] as $guid) {
        // Query the API for the full record for the item.
        $returned_item = $this->client->{$api_call[$type]}($guid, [
          'fetch-related' => TRUE,
        ]);
        // If there is a response, add it to the queue.
        if ($returned_item) {
          $this->{$manager[$type]}->getQueue()->createItem($returned_item);

          // We need to get the associated assets, so update the query.
          $query['assets'] = TRUE;
          $query['type'] = 'asset';

          if ($type == 'show') {
            $query['id'] = $guid;
            // Add the show guid to the array to avoid duplicates.
            $show_guids[] = $guid;
          }
          // If this is not a show, we need to get the show id from the item.
          else {
            if ($show_id = $returned_item->attributes->show->id) {
              // If the show is not already in the array, add it.
              if (!isset($show_guids) || !in_array($show_id,$show_guids)) {
                $query['id'] = $show_id;
                // Add the show guid to the array to avoid duplicates.
                $show_guids[] = $show_id;
              }
            }
          }

          // Get the associated assets and add them to the queue.
          if (isset($query['id'])) {
            $assets = $this->client->getChangelog($query);
            $this->processAssets($assets);
          }
        }
      }
    }

    // Set the last update key to the start time because run time
    // could be long. This means that the queue update may occasionally
    // double-up content, but it should never _miss_ any.
    $this->showManager->setLastUpdateTime($start_dt);

    return TRUE;
  }

  /**
   * Cycle through asset results from the changelog and add them to the queue.
   *
   * @param object $assets
   * Assets as returned from the changelog endpoint.
   *
   * @return void
   */
  protected function processAssets(object $assets): void {
    // Initialize new arrays to hold these results.
    $asset_guids = ['upsert' => [], 'delete' => []];
    foreach ($assets as $asset) {
      // Add items marked for deletion to an array for future processing.
      if ($asset->attributes->action === 'delete') {
        $asset_guids['delete'][$asset->id] = $asset->id;
      }
      // All other items should be created or updated.
      else {
        $asset_guids['upsert'][$asset->id] = $asset->id;
      }
    }

    if (!empty($asset_guids['upsert'])) {
      foreach ($asset_guids['upsert'] as $asset_guid) {
        // Query the API for the full record for the asset.
        $returned_asset = $this->client->getAsset($asset_guid);
        // If there is a response, add it to the queue.
        if ($returned_asset) {
          $this->assetManager->getQueue()->createItem($returned_asset);
        }
      }
    }
  }
}
