<?php

namespace Drupal\pbs_media_manager\Commands;

use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Queue\QueueWorkerManagerInterface;
use Drupal\Core\Queue\RequeueException;
use Drupal\Core\Queue\SuspendQueueException;
use Drupal\pbs_media_manager\AssetManager;
use Drupal\pbs_media_manager\EpisodeManager;
use Drupal\pbs_media_manager\EpisodeManagerNoChangelog;
use Drupal\pbs_media_manager\SeasonManager;
use Drupal\pbs_media_manager\SeasonManagerNoChangelog;
use Drupal\pbs_media_manager\ShowManager;
use Drupal\pbs_media_manager\ShowManagerNoChangelog;
use Drupal\pbs_media_manager\SpecialManager;
use Drupal\pbs_media_manager\SpecialManagerNoChangelog;
use Drush\Commands\DrushCommands;
use OpenPublicMedia\PbsMediaManager\Exception\BadRequestException;

/**
 * A Drush commandfile.
 */
class PbsCommands extends DrushCommands {

  /**
   * The PBS Media Manager show manager.
   *
   * @var \Drupal\pbs_media_manager\ShowManager
   */
  protected $showManager;

  /**
   * The PBS Media Manager show manager.
   *
   * @var \Drupal\pbs_media_manager\ShowManagerNoChangelog
   */
  protected $showManagerNoChangelog;


  /**
   * The PBS Media Manager season manager.
   *
   * @var \Drupal\pbs_media_manager\SeasonManager
   */
  protected $seasonManager;

  /**
   * The PBS Media Manager season manager.
   *
   * @var \Drupal\pbs_media_manager\SeasonManagerNoChangelog
   */
  protected $seasonManagerNoChangelog;

  /**
   * The PBS Media Manager episode manager.
   *
   * @var \Drupal\pbs_media_manager\EpisodeManager
   */
  protected $episodeManager;

  /**
   * The PBS Media Manager episode manager.
   *
   * @var \Drupal\pbs_media_manager\EpisodeManagerNoChangelog
   */
  protected $episodeManagerNoChangelog;

  /**
   * The PBS Media Manager special manager.
   *
   * @var \Drupal\pbs_media_manager\SpecialManager
   */
  protected $specialManager;

  /**
   * The PBS Media Manager special manager.
   *
   * @var \Drupal\pbs_media_manager\SpecialManagerNoChangelog
   */
  protected $specialManagerNoChangelog;

  /**
   * The PBS Media Manager asset manager.
   *
   * @var \Drupal\pbs_media_manager\AssetManager
   */
  protected $assetManager;

  /**
   * The queue worker manager.
   *
   * @var \Drupal\Core\Queue\QueueWorkerManagerInterface
   */
  protected $workerManager;

  /**
   * The queue service.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queueService;

  /**
   * Constructs a new DrushCommands object.
   *
   * @param \Drupal\pbs_media_manager\ShowManager $show_manager
   *   The PBS Media Manager show manager.
   * @param \Drupal\pbs_media_manager\SeasonManager $season_manager
   *   The PBS Media Manager season manager.
   * @param \Drupal\pbs_media_manager\EpisodeManager $episode_manager
   *   The PBS Media Manager episode manager.
   * @param \Drupal\pbs_media_manager\SpecialManager $special_manager
   *   The PBS Media Manager special manager.
   * @param \Drupal\pbs_media_manager\ShowManagerNoChangelog $show_manager_no_changelog
   *   The PBS Media Manager show manager (no changelog).
   * @param \Drupal\pbs_media_manager\SeasonManagerNoChangelog $season_manager_no_changelog
   *   The PBS Media Manager season manager (no changelog).
   * @param \Drupal\pbs_media_manager\EpisodeManagerNoChangelog $episode_manager_no_changelog
   *   The PBS Media Manager episode manager (no changelog).
   * @param \Drupal\pbs_media_manager\SpecialManagerNoChangelog $special_manager_no_changelog
   *   The PBS Media Manager special manager (no changelog).
   * @param \Drupal\pbs_media_manager\AssetManager $asset_manager
   *   The PBS Media Manager asset manager.
   * @param \Drupal\Core\Queue\QueueWorkerManagerInterface $worker_manager
   *   The queue worker manager.
   * @param \Drupal\Core\Queue\QueueFactory $queue_service
   *   The queue service.
   */
  public function __construct(ShowManager $show_manager, SeasonManager
  $season_manager, EpisodeManager $episode_manager, SpecialManager $special_manager,  ShowManagerNoChangelog $show_manager_no_changelog, SeasonManagerNoChangelog $season_manager_no_changelog, EpisodeManagerNoChangelog $episode_manager_no_changelog, SpecialManagerNoChangelog $special_manager_no_changelog, AssetManager $asset_manager, QueueWorkerManagerInterface $worker_manager, QueueFactory $queue_service) {
    parent::__construct();
    $this->showManager = $show_manager;
    $this->seasonManager = $season_manager;
    $this->episodeManager = $episode_manager;
    $this->specialManager = $special_manager;
    $this->showManagerNoChangelog = $show_manager_no_changelog;
    $this->seasonManagerNoChangelog = $season_manager_no_changelog;
    $this->episodeManagerNoChangelog = $episode_manager_no_changelog;
    $this->specialManagerNoChangelog = $special_manager_no_changelog;
    $this->assetManager = $asset_manager;
    $this->workerManager = $worker_manager;
    $this->queueService = $queue_service;
  }

  /**
   * Command to pull a single Media Manager show.
   *
   * @param string $show_id
   *   The Media Manager id of the show.
   * @param array $options
   *   Associative array of options.
   *
   * @option force Update the item even if it has not changed.
   * @option assets Include any assets associated with the item. To include
   *  the entire show tree including seasons, episodes, and specials, use
   *  pbs_media_manager:getShowAll (pbs-gsa).
   *
   * @command pbs_media_manager:getShow
   * @aliases pbs-gs, pbs:gs
   *
   */
  public function getShow(string $show_id, array $options = ['force' => FALSE,
    'assets' => FALSE]): void {
    $show = $this->showManager->getShow($show_id);
    if (is_object($show)) {
      if ($options['assets']) {
        $this->showManagerNoChangelog->addOrUpdateShow($show, $options['force']);
      }
      else {
        $this->showManager->addOrUpdateShow($show, $options['force']);
      }
    }
    else {
      $this->logger->error(dt('The show ID @id did not return any results.', [
        '@id' => $show_id,
      ]));
    }
  }

  /**
   * Command to pull a single Media Manager season.
   *
   * @param string $season_id
   *   The Media Manager id of the season.
   * @param array $options
   *   Associative array of options.
   *
   * @option force Update the item even if it has not changed.
   * @option assets Include any assets associated with the item.
   *
   * @command pbs_media_manager:getSeason
   * @aliases pbs-gn, pbs:gn
   *
   */
  public function getSeason(string $season_id, array $options = ['force' => FALSE, 'assets' => FALSE]): void {
    $season = $this->seasonManager->getSeason($season_id);
    if (is_object($season)) {
      if ($options['assets']) {
        $this->seasonManagerNoChangelog->addOrUpdateSeason($season, $options['force']);
      }
      else {
        $this->seasonManager->addOrUpdateSeason($season, $options['force']);
      }
    }
    else {
      $this->logger->error(dt('The season ID @id did not return any results.', [
        '@id' => $season_id,
      ]));
    }
  }

  /**
   * Command to pull a single Media Manager episode.
   *
   * @param string $episode_id
   *   The Media Manager id of the episode.
   * @param array $options
   *   Associative array of options.
   *
   * @option force Update the item even if it has not changed.
   * @option assets Include any assets associated with the item.
   *
   * @usage drush pbs-ge 745a2571-05fd-449e-93b6-0b8fb89e6fa1 --force
   *   Import the episode or force it to update even if it has not changed.
   *
   * @command pbs_media_manager:getEpisode
   * @aliases pbs-ge, pbs:ge
   *
   */
  public function getEpisode(string $episode_id, array $options = ['force' => FALSE,
    'assets' => FALSE]): void {
    $episode = $this->episodeManager->getEpisode($episode_id);
    if (is_object($episode)) {
      if ($options['assets']) {
        $this->episodeManagerNoChangelog->addOrUpdateEpisode($episode, $options['force']);
      }
      else {
        $this->episodeManager->addOrUpdateEpisode($episode, $options['force']);
      }
    }
    else {
      $this->logger->error(dt('The episode ID @id did not return any results.', [
        '@id' => $episode_id,
      ]));
    }
  }

  /**
   * Command to pull a single Media Manager special.
   *
   * @param string $special_id
   *   The Media Manager id of the special.
   * @param array $options
   *   Associative array of options.
   *
   * @option force Update the item even if it has not changed.
   * @option assets Include any assets associated with the item.
   *
   * @usage drush pbs-gp 398e358e-42d1-4c01-b91f-0ceb48568b16 --force
   *   Import the special or force it to update even if it has not changed.
   *
   * @command pbs_media_manager:getSpecial
   * @aliases pbs-gp, pbs:gp
   */
  public function getSpecial(string $special_id, array $options = ['force' =>
    FALSE, 'assets' => FALSE]): void {
    $special = $this->specialManager->getSpecial($special_id);
    if (is_object($special)) {
      if ($options['assets']) {
        $this->specialManagerNoChangelog->addOrUpdateSpecial($special, $options['force']);
      }
      else {
        $this->specialManager->addOrUpdateSpecial($special, $options['force']);
      }
    }
    else {
      $this->logger->error(dt('The special ID @id did not return any results.', [
        '@id' => $special_id,
      ]));
    }
  }

  /**
   * Command to pull a single Media Manager asset.
   *
   * @param string $asset_id
   *   The Media Manager id of the asset.
   * @param array $options
   *   Associative array of options.
   *
   * @option force Update the item even if it has not changed.
   *
   * @usage drush pbs-gv 82bb45da-12de-4a52-8ff9-5ee02dd30442 --force
   *   Import the asset or force it to update even if it has not changed.
   *
   * @command pbs_media_manager:getAsset
   * @aliases pbs-ga, pbs:ga
   */
  public function getAsset(string $asset_id, array $options = ['force' => FALSE]): void {
    $client = $this->assetManager->getApiClient();
    $asset = $client->getAsset($asset_id);
    if (is_object($asset)) {
      $this->assetManager->addOrUpdateAssetContent($asset, $options['force']);
    }
    else {
      $this->logger->error(dt('The asset ID @id did not return any results.', [
        '@id' => $asset_id,
      ]));
    }
  }

  /**
   * Command to add all (subscribed) shows to the Drupal queue.
   *
   * @command pbs_media_manager:getAllShows
   * @aliases pbs-gas, pbs:gas
   */
  public function getAllShows(): void {
    $this->showManager->updateQueue();
    $this->output()->writeln(dt('Process the shows with `drush queue:run pbs_media_manager.queue.shows` or run cron.'));
  }

  /**
   * Command to add all (subscribed) seasons to the Drupal queue.
   *
   * @command pbs_media_manager:getAllSeasons
   * @aliases pbs-gan
   */
  public function getAllSeasons(): void {
    $this->seasonManager->updateQueue();
    $this->output()->writeln(dt('Process the seasons with `drush queue:run pbs_media_manager.queue.seasons` or run cron.'));
  }

  /**
   * Command to add all (subscribed) episodes to the Drupal queue.
   *
   * @command pbs_media_manager:getAllEpisodes
   * @aliases pbs-gae, pbs:gae
   */
  public function getAllEpisodes(): void {
    $this->episodeManager->updateQueue();
    $this->output()->writeln(dt("Process the episodes with `drush queue:run pbs_media_manager.queue.episodes` or run cron."));
  }

  /**
   * Command to add all (subscribed) specials to the Drupal queue.
   *
   * @command pbs_media_manager:getAllSpecials
   * @aliases pbs-gap, pbs:gap
   */
  public function getAllSpecials(): void {
    $this->specialManager->updateQueue();
    $this->output()->writeln(dt("Specials are stored in the episode queue. Process the specials with `drush queue:run pbs_media_manager.queue.episodes` or run cron."));
  }

  /**
   * Command to get all content for a single show.
   *
   * @param string $show_id
   *   The Media Manager id of the show.
   * @param array $options
   *   Associative array of options.
   *
   * @option force Update the items even if they have not changed.
   *
   * @command pbs_media_manager:getShowAll
   * @aliases pbs-gsa, pbs:gsa
   */
  public function getShowAll(string $show_id, array $options = ['force' => FALSE]): void {
    $show = $this->showManager->getShow($show_id);
    if (is_object($show)) {
      $this->showManagerNoChangelog->addOrUpdateShow($show, $options['force']);
    }
    else {
      $this->logger->error(dt('The show ID @id did not return any results.', [
        '@id' => $show_id,
      ]));
    }
    $client = $this->showManager->getApiClient();
    try {
      $seasons = $client->getSeasons($show_id);
    }
    catch (BadRequestException $e) {
      // Only log show errors in ShowManager.
    }
    if (isset($seasons) && is_object($seasons)) {
      foreach ($seasons as $season) {
        if ($season_item = $client->getSeason($season->id)) {
          $this->seasonManagerNoChangelog->addOrUpdateSeason($season_item, $options['force']);
          if ($episodes = $client->getEpisodes($season->id)) {
            foreach ($episodes as $episode) {
              if ($episode_item = $client->getEpisode($episode->id)) {
                $this->episodeManagerNoChangelog->addOrUpdateEpisode($episode_item,
                  $options['force']);
              }
            }
          }
        }
      }
    }
    if ($specials = $client->getSpecials($show_id)) {
      foreach ($specials as $special) {
        if ($special_item = $client->getSpecial($special->id)) {
          $this->specialManagerNoChangelog->addOrUpdateSpecial($special_item, $options['force']);
        }
      }
    }
    $this->output()->writeln(dt("All content associated with this show has been imported."));
  }

  /**
   * Command to queue all Media Manager assets. This command has not been
   * implemented.
   *
   * @command pbs_media_manager:getAllAssets
   * @aliases pbs-gaa, pbs:gaa
   */
  public function getAllAssets() {
    $this->output()->writeln(dt("This command has not been implemented."));

    // TODO: Is this even possible in any sort of efficient way? The Assets
    // endpoint requires an ID of some sort -- unlike the other endpoints, it
    // doesn't return a list if none is provided.
  }

  /**
   * Command to process all (subscribed) shows, seasons, and episodes.
   *
   * @command pbs_media_manager:processAll
   * @aliases pbs-pa, pbs:pa
   */
  public function processAll(): void {
    $this->showManagerNoChangelog->updateQueue();
    $this->runQueue('pbs_media_manager.queue.shows');
    $this->seasonManagerNoChangelog->updateQueue();
    $this->runQueue('pbs_media_manager.queue.seasons');
    $this->episodeManagerNoChangelog->updateQueue();
    $this->runQueue('pbs_media_manager.queue.episodes');
    $this->specialManagerNoChangelog->updateQueue();
    $this->runQueue('pbs_media_manager.queue.episodes');
    $this->output()->writeln(dt("All shows, seasons, episodes, and specials have been processed."));
  }

  /**
   * Run a specific queue by name.
   *
   * @param string $name
   *   The name of the queue to run.
   *
   * @command pbs:rq
   * @aliases pbs-rq
   */
  public function runQueue($name) {
    $worker = $this->workerManager->createInstance($name);
    $queue = $this->queueService->get($name);
    $start = microtime(TRUE);
    $count = 0;

    while ($item = $queue->claimItem()) {
      try {
        $this->logger()->info(dt('Processing item @id from @name queue.', [
          '@name' => $name,
          '@id' => $item->item_id,
        ]));
        $worker->processItem($item->data);
        $queue->deleteItem($item);
        $count++;
      }
      catch (RequeueException $e) {
        // The worker requested the task to be immediately requeued.
        $queue->releaseItem($item);
      }
      catch (SuspendQueueException $e) {
        // If the worker indicates there is a problem with the whole queue,
        // release the item.
        $queue->releaseItem($item);
        throw new \Exception($e->getMessage());
      }
    }
    $elapsed = microtime(TRUE) - $start;
    $this->logger()->success(dt('Processed @count items from the @name queue in @elapsed sec.', [
      '@count' => $count,
      '@name' => $name,
      '@elapsed' => round($elapsed, 2),
    ]));
  }

}
