<?php

namespace Drupal\pbs_media_manager;

use DateTime;
use DateTimeZone;
use Drupal;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\State\StateInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Drupal\node\NodeInterface;
use Drupal\taxonomy\Entity\Term;
use Exception;
use OpenPublicMedia\PbsMediaManager\Exception\BadRequestException;
use stdClass;

/**
 * Class EpisodeManager.
 *
 * @package Drupal\pbs_media_manager
 */
class EpisodeManager extends ApiContentManagerBase {

  /**
   * Show manager.
   *
   * @var \Drupal\pbs_media_manager\ShowManager
   */
  protected ShowManager $showManager;

  /**
   * {@inheritdoc}
   */
  public static function getEntityTypeId(): string {
    return 'node';
  }

  /**
   * {@inheritdoc}
   */
  public function getBundleId(): string {
    return $this->config->get('episodes.drupal_episode_content');
  }

  /**
   * {@inheritdoc}
   */
  public static function getQueueName(): string {
    return 'pbs_media_manager.queue.episodes';
  }

  /**
   * {@inheritdoc}
   */
  public function getLastUpdateTime(): DateTime {
    return new DateTime();
  }

  /**
   * {@inheritdoc}
   */
  public function setLastUpdateTime(DateTime $time): void {
  }

  /**
   * {@inheritdoc}
   */
  public function resetLastUpdateTime(): void {
  }

  /**
   * {@inheritdoc}
   */
  public static function getAutoUpdateConfigName(): string {
    return '';
  }

  /**
   * {@inheritdoc}
   */
  public static function getAutoUpdateIntervalConfigName(): string {
    return '';
  }

  /**
   * ApiContentManagerBase constructor.
   *
   * {@inheritdoc}
   */
  public function __construct(
    ApiClient $client,
    EntityTypeManagerInterface $entity_type_manager,
    ModuleHandlerInterface $module_handler,
    ConfigFactoryInterface $config_factory,
    LoggerChannelInterface $logger,
    QueueFactory $queue_factory,
    StateInterface $state,
    FileSystemInterface $file_system,
    ShowManager $show_manager
  ) {
    parent::__construct($client, $entity_type_manager, $module_handler,
      $config_factory, $logger, $queue_factory, $state, $file_system);
    $this->showManager = $show_manager;
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   */
  public function updateQueue(DateTime $since = NULL): bool {
    $logging = $this->config->get('logging.operations');

    // Initialize counter for tracking items added to queue.
    $count = 0;

    // Check to see if we're subscribed to specific shows.
    $subscription = FALSE;
    $show_vocabulary = $this->config->get('shows.show_vocabulary');
    if (!empty($show_vocabulary) && $show_vocabulary != 'unused') {
      $subscription = TRUE;
      // Get the shows.
      $show_guids = $this->showManager->getSubscribedShows($show_vocabulary);
    }

    // Check to see if we're using the changelog.
    $changelog = $this->config->get('shows.queue.changelog');
    if ($changelog) {
      if (empty($since)) {
        $since = $this->getLastUpdateTime();
      }

      // Get UTC datetime for "since" argument to changelog endpoint.
      $utc = new DateTimeZone('UTC');
      $since_utc = clone($since);
      $since_utc->setTimezone($utc);

      // Build the query to get the updated episodes from the changelog.
      $query = [
        'type' => 'episode',
        'since' => $since_utc->format(self::API_DATETIME_FORMAT),
        'sort' => '-timestamp',
      ];

      // Get the changelog.
      $changelog_result = $this->client->getChangelog($query);
      // Divide into upsert vs. delete.
      $changelog_result = $this->sortChangelog($changelog_result);

      // Cycle through the created/updated objects to determine if we have
      // access.
      foreach ($changelog_result['upsert'] as $guid) {
        // Query the API for the full record for the item.
        if ($returned_item = $this->client->getEpisode($guid, ['fetch-related' =>
          TRUE,
        ])) {
          // If subscribed to specific shows, check that this episode belongs to
          // one of those shows before adding it to the queue.
          if ($subscription) {
            if (isset($show_guids[$returned_item->attributes->show->id])) {
              $this->getQueue()->createItem($returned_item);
              $count++;
            }
          }
          // Otherwise, add all returned items to the queue.
          else {
            $this->getQueue()->createItem($returned_item);
            $count++;
          }
        }
      }
      // Cycle through the deleted episodes and delete the associated content
      // entities.
      foreach ($changelog_result['delete'] as $guid) {
        if ($node = $this->getNodeByGuid($guid, $this->getBundleId(), 'episode')) {
          if (!$logging['changed']) {
            $this->logger->notice($this->t('The episode @title will be deleted', [
              '@title' => $node->getTitle(),
            ]));
          }
          $node->delete();
        };
      }
    }
    // If we're not using the changelog, we need to cycle through the Episode
    // endpoint to add items to the queue.
    else {
      // If only subscribed to certain shows, only add episodes for those
      // shows to the queue.
      if ($subscription) {
        if (!empty($show_guids)) {
          foreach ($show_guids as $guid => $name) {
            $count += $this->updateEpisodesQueue($guid, $name);
          }
        }
      }
      // Otherwise, add all episodes to the queue.
      else {
        $shows = $this->client->getShows(['fetch-related' => FALSE]);
        foreach ($shows as $show) {
          $count += $this->updateEpisodesQueue($show->id, $show->attributes->title);

        }
      }
    }

    if (!$logging['queue']) {
      $this->logger->notice('{count} total episodes have been added to the queue.', ['count' => $count,]);
    }

    return TRUE;
  }

  /**
   * Gets an API response for a single Episode.
   *
   * @param string $guid
   *   ID of the Episode to get.
   *
   * @return object|null
   *   Episode item data from the API or NULL.
   */
  public function getEpisode(string $guid): ?stdClass {
    return $this->client->getEpisode($guid);
  }

  /**
   * Get Episode nodes with optional properties.
   *
   * @param array $properties
   *   Properties to filter Episode nodes.
   *
   * @return \Drupal\node\NodeInterface[]
   *   Episodes nodes.
   */
  public function getEpisodeNodes(array $properties = []): array {
    try {
      $definition = $this->entityTypeManager->getDefinition(self::getEntityTypeId());
      $storage = $this->entityTypeManager->getStorage(self::getEntityTypeId());
      $nodes = $storage->loadByProperties([
        $definition->getKey('bundle') => $this->getBundleId(),
      ] + $properties);
    }
    catch (Exception $e) {
      // Let NULL fall through.
      $nodes = [];
    }

    return $nodes;
  }

  /**
   * Gets a Episode node by TMS ID.
   *
   * @param string $id
   *   TMS ID to query with.
   *
   * @return \Drupal\node\NodeInterface|null
   *   Episode node with TMS ID or NULL if none found.
   */
  public function getEpisodeNodeByTmsId(string $id): ?NodeInterface {
    if ($tms_id_field = $this->config->get('episodes.mappings.tms_id')) {
      $nodes = $this->getEpisodeNodes([$tms_id_field => $id]);

      $node = NULL;
      if (!empty($nodes)) {
        $node = reset($nodes);
        if (count($nodes) > 1) {
          $this->logger->error('Multiple nodes found for Media Manager
            TMS ID {id}. Node IDs found: {nid_list}. Using node {nid}.', [
              'id' => $id,
              'nid_list' => implode(', ', array_keys($nodes)),
              'nid' => $node->id(),
            ]);
        }
      }
    }
    else {
      $this->logger->error('Mapping the episode TMS ID field is required.');
    }
    return $node;
  }

  /**
   * Gets a Episode node by slug.
   *
   * @param string $slug
   *   Slug to query with.
   *
   * @return \Drupal\node\NodeInterface|null
   *   Episode node with the slug or NULL if none found.
   */
  public function getEpisodeNodeBySlug(string $slug): ?NodeInterface {
    $node = NULL;
    if ($slug_field = $this->config->get('episodes.mappings.slug')) {
      $nodes = $this->getEpisodeNodes([$slug_field => $slug]);
      if (!empty($nodes)) {
        $node = reset($nodes);
      }
    }
    return $node;
  }

  /**
   * Adds or updates an Episode node based on API data.
   *
   * @param object $item
   *   An API response object for the Episode.
   * @param bool $force
   *   Whether to "force" the update. If FALSE (default) and a current
   *   episode content node is found with a matching or newer "last_updated"
   *   value, the existing node will not be updated. If TRUE, "last_updated" is
   *   ignored.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Exception
   */
  public function addOrUpdateEpisode(object $item, $force = FALSE): void {
    $node = $this->getOrCreateNode($item->id, $this->getBundleId(), 'episodes');
    $mappings = $this->config->get('episodes.mappings');
    $logging = $this->config->get('logging.operations');
    $operation = $this->getOperation($item, $node, $mappings, $force);

    if ($operation == 'configured') {
      $this->logger->error($this->t('Please map the title, id, and updated_at fields for episodes.'));
      return;
    }
    // Do not consider videos for ignored episodes.
    if ($operation == 'ignored') {
      if (!$logging['ignored']) {
        $this->logger->debug($this->t('The episode @id was ignored due to a hook_pbs_media_manager_operation_alter() implementation.', [
          '@id' => $item->id,
        ]));
      }
      return;
    }

    $attributes = $item->attributes;
    if (empty($attributes->title)) {
      return;
    }
    $node->setTitle($attributes->title);

    // We need to save the node before creating the videos because episodes
    // reference videos and videos reference episodes. Therefore, we also have
    // to set the title of the episode. Additionally, we will also need the
    // guid because addOrUpdateAssetContent() looks up nodes by guid.
    if ($node->isNew()) {
      if (!empty($item->id) && $mappings['id'] != 'unused') {
        $node->set($mappings['id'], $item->id);
      }
      $node->save();
    }

    if ($operation == 'skipped') {
      if (!$logging['skipped']) {
        $this->logger->debug($this->t('The episode @title has not changed', [
          '@title' => $node->getTitle(),
        ]));
      }
      return;
    }

    $type = $mappings['type'];
    $episode_type_vocab = $this->config->get('episodes.references.episode_type_vocabularly');
    if (!empty($item->type) && !empty($type) && $type != 'unused') {
      $taxonomy_storage = $this->entityTypeManager->getStorage('taxonomy_term');
      if ($term = $taxonomy_storage->loadByProperties([
        'name' => $item->type,
        'vid' => $episode_type_vocab,
      ])) {
        $term = reset($term);
      }
      else {
        $term = Term::create([
          'name' => $item->type,
          'vid' => $episode_type_vocab,
        ]);
        $term->save();
      }
      $node->set($type, ['target_id' => $term->id()]);
    }
    $other_attributes = [
      'tms_id',
      'ordinal',
      'description_long',
      'description_short',
      'nola',
      'premiered_on',
      'slug',
      'encored_on',
      'language',
    ];
    foreach ($other_attributes as $attribute) {
      $attribute_field = $mappings[$attribute];
      if (!empty($attributes->{$attribute}) && !empty($attribute_field) && $attribute_field != 'unused') {
        $node->set($attribute_field, $attributes->{$attribute});
      }
    }

    $updated_at = self::getLatestUpdatedAt($item);
    if ($mappings['updated_at'] != 'unused') {
      $node->set(
        $mappings['updated_at'],
        $updated_at->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT)
      );
    }

    // Process images.
    if (!empty($attributes->assets)) {
      foreach($attributes->assets as $asset) {
        // There can be multiple "related" assets.
        // First check for a full length asset, then fall back to the first
        // image available.
        if (!empty($asset->attributes->images)) {
          if (isset($asset->attributes->object_type) && $asset->attributes->object_type == 'full_length') {
            $images = $asset->attributes->images;
            break;
          }
          elseif (empty($images)) {
            $images = $asset->attributes->images;
          }
        }
      }
    }

    if (isset($images) && is_array($images)) {
      foreach ($images as $image) {
        if (!empty($mappings[$image->profile]) && $image_field = $mappings[$image->profile]) {
          if ($image_field != 'unused') {
            // Check to see if this item has an image already.
            $media_image = NULL;
            if ($node->get($image_field)->target_id) {
              // Get the existing image entity.
              $image_id = $node->get($image_field)->target_id;
              $media_storage = $this->entityTypeManager->getStorage('media');
              $media_image = $media_storage->load($image_id);
            }
            if ($media_id = $this->addOrUpdateMediaImage(
              $image,
              $attributes->title . ": " . $image->profile,
              'episode',
              $media_image
            )) {
              $node->set($image_field, ['target_id' => $media_id]);
            }
          }
        }
      }
    }

    // Add the show reference, if configured.
    $show_ref_field = $this->config->get('episodes.references.show');
    $show_guid = $attributes->show->id;
    $show_bundle = $this->config->get('shows.drupal_show_content');
    if (!empty($show_guid) && !empty($show_bundle) && $show_ref_field != 'unused') {
      $show_node = $this->getNodeByGuid($show_guid, $show_bundle, 'shows');
      if (!empty($show_node) && !empty($show_node->id())) {
        $node->set($show_ref_field, ['target_id' => $show_node->id()]);
      }
    }

    // Add the season reference, if configured.
    $season_ref_field = $this->config->get('episodes.references.season');
    $season_guid = NULL;
    if (isset($attributes->season->id)) {
      $season_guid = $attributes->season->id;
    }
    $season_bundle = $this->config->get('seasons.drupal_season_content');
    if (!empty($season_guid) && !empty($season_bundle) && $season_ref_field != 'unused') {
      $season_node = $this->getNodeByGuid($season_guid, $season_bundle, 'seasons');
      if (!empty($season_node) && !empty($season_node->id())) {
        $node->set($season_ref_field, ['target_id' => $season_node->id()]);
      }
    }

    // Save the node.
    $node->save();

    if (!empty($node->getTitle()) & !empty($operation && !$logging['changed'])) {
      $this->logger->notice($this->t('Episode @title has been @operation', [
        '@title' => $node->getTitle(),
        '@operation' => $operation,
      ]));
    }
  }


}
