<?php

namespace Drupal\pbs_media_manager;

use DateTime;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\State\StateInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Drupal\taxonomy\Entity\Term;
use OpenPublicMedia\PbsMediaManager\Exception\BadRequestException;

class EpisodeManagerNoChangelog extends EpisodeManager {

  /**
   * Asset manager.
   *
   * @var \Drupal\pbs_media_manager\AssetManager
   */
  protected AssetManager $assetManager;

  /**
   * ApiContentManagerBase constructor.
   *
   * {@inheritdoc}
   */
  public function __construct(
    ApiClient $client,
    EntityTypeManagerInterface $entity_type_manager,
    ModuleHandlerInterface $module_handler,
    ConfigFactoryInterface $config_factory,
    LoggerChannelInterface $logger,
    QueueFactory $queue_factory,
    StateInterface $state,
    FileSystemInterface $file_system,
    ShowManager $show_manager,
    AssetManager $asset_manager
  ) {
    parent::__construct($client, $entity_type_manager, $module_handler,
      $config_factory, $logger, $queue_factory, $state, $file_system,
      $show_manager);
    $this->assetManager = $asset_manager;
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   */
  public function updateQueue(DateTime $since = NULL): bool {
    $logging = $this->config->get('logging.operations');

    // Initialize counter for tracking items added to queue.
    $count = 0;

    // Check to see if we're subscribed to specific shows.
    $subscription = FALSE;
    $show_vocabulary = $this->config->get('shows.show_vocabulary');
    if (!empty($show_vocabulary) && $show_vocabulary != 'unused') {
      $subscription = TRUE;
      // Get the shows.
      $show_guids = $this->showManager->getSubscribedShows($show_vocabulary);
    }

    // Since we're not using the changelog, we need to cycle through the Episode
    // endpoint to add items to the queue.

    // If only subscribed to certain shows, only add episodes for those
    // shows to the queue.
    if ($subscription) {
      if (!empty($show_guids)) {
        foreach ($show_guids as $guid => $name) {
          $count += $this->updateEpisodesQueue($guid, $name);
        }
      }
    }
    // Otherwise, add all episodes to the queue.
    else {
      $shows = $this->client->getShows(['fetch-related' => FALSE]);
      foreach ($shows as $show) {
        $count += $this->updateEpisodesQueue($show->id, $show->attributes->title);

      }
    }


    if (!$logging['queue']) {
      $this->logger->notice('{count} total episodes have been added to the queue.', ['count' => $count,]);
    }

    return TRUE;
  }

  /**
   * Adds or updates an Episode node based on API data.
   *
   * @param object $item
   *   An API response object for the Episode.
   * @param bool $force
   *   Whether to "force" the update. If FALSE (default) and a current
   *   episode content node is found with a matching or newer "last_updated"
   *   value, the existing node will not be updated. If TRUE, "last_updated" is
   *   ignored.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Exception
   */
  public function addOrUpdateEpisode(object $item, $force = FALSE): void {
    $node = $this->getOrCreateNode($item->id, $this->getBundleId(), 'episodes');
    $mappings = $this->config->get('episodes.mappings');
    $logging = $this->config->get('logging.operations');
    $operation = $this->getOperation($item, $node, $mappings, $force);

    if ($operation == 'configured') {
      $this->logger->error($this->t('Please map the title, id, and updated_at fields for episodes.'));
      return;
    }
    // Do not consider videos for ignored episodes.
    if ($operation == 'ignored') {
      if (!$logging['ignored']) {
        $this->logger->debug($this->t('The episode @id was ignored due to a hook_pbs_media_manager_operation_alter() implementation.', [
          '@id' => $item->id,
        ]));
      }
      return;
    }

    $attributes = $item->attributes;
    if (empty($attributes->title)) {
      return;
    }
    $node->setTitle($attributes->title);

    // We need to save the node before creating the videos because episodes
    // reference videos and videos reference episodes. Therefore, we also have
    // to set the title of the episode. Additionally, we will also need the
    // guid because addOrUpdateAssetContent() looks up nodes by guid.
    if ($node->isNew()) {
      if (!empty($item->id) && $mappings['id'] != 'unused') {
        $node->set($mappings['id'], $item->id);
      }
      $node->save();
    }

    // Since we are NOT using the changelog, we must check all the assets for
    // the episode to determine if they have been updated.
    if ($video_field = $mappings['video']) {
      if (!empty($video_field) && $video_field !== 'unused' && !empty($attributes->assets)) {
        $videos = $this->client->get('episodes/' . $item->id . "/assets");
        $video_drupal_ids = [];
        foreach ($videos as $video) {
          if ($media_video = $this->assetManager->addOrUpdateAssetContent
          ($video,
            $force)) {
            $video_drupal_ids[] = ['target_id' => $media_video->id()];
          }
        }
        $node->set($video_field, $video_drupal_ids);
        // Even if the Episode has not changed, we added or updated videos, so
        // save those to the Episode.
        $node->save();
      }
    }



    if ($operation == 'skipped') {
      if (!$logging['skipped']) {
        $this->logger->debug($this->t('The episode @title has not changed', [
          '@title' => $node->getTitle(),
        ]));
      }
      return;
    }

    $type = $mappings['type'];
    $episode_type_vocab = $this->config->get('episodes.references.episode_type_vocabularly');
    if (!empty($item->type) && !empty($type) && $type != 'unused') {
      $taxonomy_storage = $this->entityTypeManager->getStorage('taxonomy_term');
      if ($term = $taxonomy_storage->loadByProperties([
        'name' => $item->type,
        'vid' => $episode_type_vocab,
      ])) {
        $term = reset($term);
      }
      else {
        $term = Term::create([
          'name' => $item->type,
          'vid' => $episode_type_vocab,
        ]);
        $term->save();
      }
      $node->set($type, ['target_id' => $term->id()]);
    }
    $other_attributes = [
      'tms_id',
      'ordinal',
      'description_long',
      'description_short',
      'nola',
      'premiered_on',
      'slug',
      'encored_on',
      'language',
    ];
    foreach ($other_attributes as $attribute) {
      $attribute_field = $mappings[$attribute];
      if (!empty($attributes->{$attribute}) && !empty($attribute_field) && $attribute_field != 'unused') {
        $node->set($attribute_field, $attributes->{$attribute});
      }
    }

    $updated_at = self::getLatestUpdatedAt($item);
    if ($mappings['updated_at'] != 'unused') {
      $node->set(
        $mappings['updated_at'],
        $updated_at->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT)
      );
    }

    // Process images.
    if (!empty($attributes->assets)) {
      foreach($attributes->assets as $asset) {
        // There can be multiple "related" assets.
        // First check for a full length asset, then fall back to the first
        // image available.
        if (!empty($asset->attributes->images)) {
          if (isset($asset->attributes->object_type) && $asset->attributes->object_type == 'full_length') {
            $images = $asset->attributes->images;
            break;
          }
          elseif (empty($images)) {
            $images = $asset->attributes->images;
          }
        }
      }
    }

    if (isset($images) && is_array($images)) {
      foreach ($images as $image) {
        if (!empty($mappings[$image->profile]) && $image_field = $mappings[$image->profile]) {
          if ($image_field != 'unused') {
            // Check to see if this item has an image already.
            $media_image = NULL;
            if ($node->get($image_field)->target_id) {
              // Get the existing image entity.
              $image_id = $node->get($image_field)->target_id;
              $media_storage = $this->entityTypeManager->getStorage('media');
              $media_image = $media_storage->load($image_id);
            }
            if ($media_id = $this->addOrUpdateMediaImage(
              $image,
              $attributes->title . ": " . $image->profile,
              'episode',
              $media_image
            )) {
              $node->set($image_field, ['target_id' => $media_id]);
            }
          }
        }
      }
    }

    // Add the show reference, if configured.
    $show_ref_field = $this->config->get('episodes.references.show');
    $show_guid = $attributes->show->id;
    $show_bundle = $this->config->get('shows.drupal_show_content');
    if (!empty($show_guid) && !empty($show_bundle) && $show_ref_field != 'unused') {
      $show_node = $this->getNodeByGuid($show_guid, $show_bundle, 'shows');
      if (!empty($show_node) && !empty($show_node->id())) {
        $node->set($show_ref_field, ['target_id' => $show_node->id()]);
      }
    }

    // Add the season reference, if configured.
    $season_ref_field = $this->config->get('episodes.references.season');
    $season_guid = NULL;
    if (isset($attributes->season->id)) {
      $season_guid = $attributes->season->id;
    }
    $season_bundle = $this->config->get('seasons.drupal_season_content');
    if (!empty($season_guid) && !empty($season_bundle) && $season_ref_field != 'unused') {
      $season_node = $this->getNodeByGuid($season_guid, $season_bundle, 'seasons');
      if (!empty($season_node) && !empty($season_node->id())) {
        $node->set($season_ref_field, ['target_id' => $season_node->id()]);
      }
    }

    // Save the node.
    $node->save();

    if (!empty($node->getTitle()) & !empty($operation && !$logging['changed'])) {
      $this->logger->notice($this->t('Episode @title has been @operation', [
        '@title' => $node->getTitle(),
        '@operation' => $operation,
      ]));
    }
  }

  /**
   * Query the Episodes endpoint to add items to the queue.
   *
   * @param string $show_guid
   *   A GUID for a specific show.
   * @param string $show_name
   *   The name of the show.
   *
   * @return int
   *   Returns the total number of episodes added per run for logging purposes.
   * @throws \OpenPublicMedia\PbsMediaManager\Exception\BadRequestException
   */
  protected function updateEpisodesQueue(string $show_guid, string $show_name): int {
    $logging = $this->config->get('logging.operations');
    // Initialize counters for tracking items added to queue.
    $count = 0;
    $count_per_show = 0;

    // We have to drill down through each season, so get all seasons per show
    // from the API.
    try {
      $seasons = $this->client->getSeasons($show_guid);
    }
    catch (BadRequestException $e) {
      // Only log show errors in ShowManager.
    }

    // Cycle through the response and get all episodes for each season.
    if (isset($seasons) && is_object($seasons)) {
      foreach ($seasons as $season) {
        $episodes = $this->client->getEpisodes($season->id);
        foreach ($episodes as $episode) {
          $episode_item = $this->client->getEpisode($episode->id);
          if (is_object($episode_item)) {
            $this->getQueue()->createItem($episode_item);
            $count++;
            $count_per_show++;
            // Give an incremental update every 50 episodes to avoid timeouts for a show like PBS NewsHour with many episodes.
            if ($count_per_show % 50 === 0 && !$logging['queue']) {
              $this->logger->debug('{show} update: {count} episodes added to the queue.',
                [
                  'count' => $count_per_show,
                  'show' => $show_name
                ]);
            }
          }
        }
        if ($count_per_show && !$logging['queue']) {
          $this->logger->debug('{count} episodes have been added to the queue for show {show}.',
            [
              'count' => $count_per_show,
              'show' => $show_name
            ]);
        }
      }
    }
    if ($count_per_show && !$logging['queue']) {
      $this->logger->debug('{count} episodes have been added to the queue for show {show}.', [
        'count' => $count_per_show,
        'show' => $show_name
      ]);
    }

    return $count;

  }


}
