<?php

namespace Drupal\pbs_media_manager;

use DateTime;
use DateTimeZone;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\State\StateInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Exception;
use OpenPublicMedia\PbsMediaManager\Exception\BadRequestException;
use stdClass;

/**
 * Class SeasonManager.
 *
 * @package Drupal\pbs_media_manager
 */
class SeasonManager extends ApiContentManagerBase {

  /**
   * State key for the last update DateTime.
   */
  const LAST_UPDATE_KEY = 'pbs_media_manager.shows.last_update';

  /**
   * Show manager.
   *
   * @var \Drupal\pbs_media_manager\ShowManager
   */
  protected ShowManager $showManager;


  /**
   * {@inheritdoc}
   */
  public static function getEntityTypeId(): string {
    return 'node';
  }

  /**
   * {@inheritdoc}
   */
  public function getBundleId(): string {
    return $this->config->get('seasons.drupal_season_content');
  }

  /**
   * {@inheritdoc}
   */
  public static function getQueueName(): string {
    return 'pbs_media_manager.queue.seasons';
  }

  /**
   * {@inheritdoc}
   */
  public function getLastUpdateTime(): DateTime {
    return $this->state->get(
      self::LAST_UPDATE_KEY,
      new DateTime('@1')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function setLastUpdateTime(DateTime $time): void {
    $this->state->set(self::LAST_UPDATE_KEY, $time);
  }

  /**
   * {@inheritdoc}
   */
  public function resetLastUpdateTime(): void {
    $this->state->delete(self::LAST_UPDATE_KEY);
  }

  /**
   * {@inheritdoc}
   */
  public static function getAutoUpdateConfigName(): string {
    return 'season.queue.autoupdate';
  }

  /**
   * {@inheritdoc}
   */
  public static function getAutoUpdateIntervalConfigName(): string {
    return 'seasons.queue.autoupdate_interval';
  }

  /**
   * ApiContentManagerBase constructor.
   *
   * {@inheritdoc}
   */
  public function __construct(
    ApiClient $client,
    EntityTypeManagerInterface $entity_type_manager,
    ModuleHandlerInterface $module_handler,
    ConfigFactoryInterface $config_factory,
    LoggerChannelInterface $logger,
    QueueFactory $queue_factory,
    StateInterface $state,
    FileSystemInterface $file_system,
    ShowManager $show_manager
  ) {
    parent::__construct($client, $entity_type_manager, $module_handler,
      $config_factory, $logger, $queue_factory, $state, $file_system);
    $this->showManager = $show_manager;
  }


  /**
   * {@inheritdoc}
   * @throws \OpenPublicMedia\PbsMediaManager\Exception\BadRequestException
   */
  public function updateQueue(DateTime $since = NULL): bool {
    $logging = $this->config->get('logging.operations');

    // Initialize counter.
    $count = 0;

    // Check to see if we're subscribed to specific shows.
    $subscription = FALSE;
    $show_vocabulary = $this->config->get('shows.show_vocabulary');
    if (!empty($show_vocabulary) && $show_vocabulary != 'unused') {
      $subscription = TRUE;
      // Get the shows.
      $show_guids = $this->showManager->getSubscribedShows($show_vocabulary);
    }

    if (empty($since)) {
      $since = $this->getLastUpdateTime();
    }
    // Get UTC datetime for "since" argument to changelog endpoint.
    $utc = new DateTimeZone('UTC');
    $since_utc = clone($since);
    $since_utc->setTimezone($utc);

    // Get the updated seasons from the changelog.
    $query = [
      'type' => 'season',
      'since' => $since_utc->format(self::API_DATETIME_FORMAT),
      'sort' => '-timestamp',
    ];

    // Get the changelog.
    $changelog_result = $this->client->getChangelog($query);
    $changelog_result = $this->sortChangelog($changelog_result);

    // Cycle through the created/updated objects to determine if we have
    // access.
    foreach ($changelog_result['upsert'] as $guid) {
      // Query the API for the full record for the item.
      if ($returned_item = $this->client->getSeason($guid, ['fetch-related' => TRUE,
      ])) {
        // If subscribed to specific shows, check that this season belongs
        // to one of those shows before adding it to the queue.
        if ($subscription) {
          if (isset($show_guids[$returned_item->attributes->show->id])) {
            $this->getQueue()->createItem($returned_item);
            $count++;
          }
        }
        // Otherwise, add all returned items to the queue.
        else {
          $this->getQueue()->createItem($returned_item);
          $count++;
        }
      }
    }
    // Cycle through the deleted seasons and delete the associated content
    // entities.
    foreach ($changelog_result['delete'] as $guid) {
      if ($node = $this->getNodeByGuid($guid, $this->getBundleId(),
        'season')) {
        if (!$logging['changed']) {
          $this->logger->notice($this->t('The season @title will be deleted', [
            '@title' => $node->getTitle(),
          ]));
        }
        $node->delete();
      };
    }


    if (!$logging['queue']) {
      $this->logger->notice('{count} total seasons have been added to the queue.', ['count' => $count,]);
    }

    return TRUE;
  }

  /**
   * Gets an API response for a single season.
   *
   * @param string $guid
   *   ID of the season to get.
   *
   * @return object|null
   *   season item data from the API or NULL.
   */
  public function getSeason(string $guid): ?stdClass {
    return $this->client->getSeason($guid);
  }

  /**
   * Get season nodes with optional properties.
   *
   * @param array $properties
   *   Properties to filter season nodes.
   *
   * @return \Drupal\node\NodeInterface[]
   *   seasons nodes.
   */
  public function getSeasonNodes(array $properties = []): array {
    try {
      $definition = $this->entityTypeManager->getDefinition(self::getEntityTypeId());
      $storage = $this->entityTypeManager->getStorage(self::getEntityTypeId());
      $nodes = $storage->loadByProperties([
        $definition->getKey('bundle') => $this->getBundleId(),
      ] + $properties);
    }
    catch (Exception $e) {
      // Let NULL fall through.
      $nodes = [];
    }

    return $nodes;
  }

  /**
   * Adds or updates a season node based on API data.
   *
   * @param object $item
   *   An API response object for the season.
   * @param bool $force
   *   Whether to "force" the update. If FALSE (default) and a current
   *   season content node is found with a matching or newer "last_updated"
   *   value, the existing node will not be updated. If TRUE, "last_updated" is
   *   ignored.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Exception
   */
  public function addOrUpdateSeason(object $item, $force = FALSE): void {

    $mappings = $this->config->get('seasons.mappings');
    $node = $this->getOrCreateNode($item->id, $this->getBundleId(), 'seasons');

    $operation = $this->getOperation($item, $node, $mappings, $force);
    $logging = $this->config->get('logging.operations');

    if ($operation == 'configured') {
      $this->logger->error($this->t('Please map the title, id, and updated_at fields for seasons.'));
      return;
    }
    if ($operation == 'skipped') {
      if (!$logging['skipped']) {
        $this->logger->debug($this->t('The season @title has not changed', [
          '@title' => $node->getTitle(),
        ]));
      }
      return;
    }

    // We need to save the node before creating the videos because seasons
    // reference videos and videos reference seasons. Therefore, we also have
    // to set the title of the season. Additionally, we need the guid because
    // addOrUpdateAssetContent() looks up nodes by guid.
    $attributes = $item->attributes;
    $season_number = $attributes->ordinal ?: '';
    $show_title = $attributes->show->attributes->title ?: '';
    if (!empty($attributes->title)) {
      $season_title = $attributes->title;
    }
    elseif (!empty($season_number) && !empty($show_title)) {
      $season_title = $show_title . ": Season " . $season_number;
    }
    else {
      $this->logger->error($this->t('The season ID @id is missing a title, season number, and/or associated show title.', [
        '@id' => $item->id,
      ]));
      return;
    }
    $node->setTitle($season_title);
    if (!empty($item->id) && $mappings['id'] != 'unused') {
      $node->set($mappings['id'], $item->id);
    }
    $node->save();

    if ($operation == 'ignored') {
      if (!$logging['ignored']) {
        $this->logger->debug($this->t('The season @id was ignored due to a hook_pbs_media_manager_operation_alter() implementation.', [
          '@id' => $item->id,
        ]));
      }
      return;
    }

    if (!empty($item->id) && $mappings['id'] != 'unused') {
      $node->set($mappings['id'], $item->id);
    }
    if (!empty($attributes->ordinal) && $mappings['ordinal'] != 'unused') {
      $node->set($mappings['ordinal'], $attributes->ordinal);
    }
    if (!empty($attributes->description_long) &&
      $mappings['description_long'] != 'unused') {
      $node->set($mappings['description_long'], $attributes->description_long);
    }
    if (!empty($attributes->description_short) &&
      $mappings['description_short'] != 'unused') {
      $node->set($mappings['description_short'], $attributes->description_short);
    }

    $updated_at = self::getLatestUpdatedAt($item);
    $node->set(
      $mappings['updated_at'],
      $updated_at->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT)
    );

    // Add the show reference, if configured.
    $show_ref_field = $this->config->get('seasons.references.show');
    $show_guid = $attributes->show->id;
    $show_bundle = $this->config->get('shows.drupal_show_content');
    if (!empty($show_guid) && !empty($show_bundle) && $show_ref_field != 'unused') {
      $show_node = $this->getNodeByGuid($show_guid, $show_bundle, 'shows');
      if (!empty($show_node) && !empty($show_node->id())) {
        $node->set($show_ref_field, ['target_id' => $show_node->id()]);
      }
    }

    $node->save();

    if (!empty($node->getTitle()) & !empty($operation && !$logging['changed'])) {
      $this->logger->notice($this->t('Season @title has been @operation', [
        '@title' => $node->getTitle(),
        '@operation' => $operation,
      ]));
    }
  }


}
