<?php

namespace Drupal\pbs_media_manager;

use DateTime;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\State\StateInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use OpenPublicMedia\PbsMediaManager\Exception\BadRequestException;

class SeasonManagerNoChangelog extends SeasonManager {
  /**
   * Asset manager.
   *
   * @var \Drupal\pbs_media_manager\AssetManager
   */
  protected AssetManager $assetManager;

  /**
   * ApiContentManagerBase constructor.
   *
   * {@inheritdoc}
   */
  public function __construct(
    ApiClient $client,
    EntityTypeManagerInterface $entity_type_manager,
    ModuleHandlerInterface $module_handler,
    ConfigFactoryInterface $config_factory,
    LoggerChannelInterface $logger,
    QueueFactory $queue_factory,
    StateInterface $state,
    FileSystemInterface $file_system,
    ShowManager $show_manager,
    AssetManager $asset_manager
  ) {
    parent::__construct($client, $entity_type_manager, $module_handler,
      $config_factory, $logger, $queue_factory, $state, $file_system,
      $show_manager);
    $this->assetManager = $asset_manager;
  }

  /**
   * {@inheritdoc}
   * @throws \OpenPublicMedia\PbsMediaManager\Exception\BadRequestException
   */
  public function updateQueue(DateTime $since = NULL): bool {
    $logging = $this->config->get('logging.operations');

    // Initialize counter.
    $count = 0;

    // Check to see if we're subscribed to specific shows.
    $subscription = FALSE;
    $show_vocabulary = $this->config->get('shows.show_vocabulary');
    if (!empty($show_vocabulary) && $show_vocabulary != 'unused') {
      $subscription = TRUE;
      // Get the shows.
      $show_guids = $this->showManager->getSubscribedShows($show_vocabulary);
    }

    // Since we're not using the changelog, we need to cycle through the Season
    // endpoint to add items to the queue.

    // If only subscribed to certain shows, only add seasons for those shows to
    // the queue.
    if ($subscription) {
      if (!empty($show_guids)) {
        foreach ($show_guids as $guid => $name) {
          $count += $this->updateSeasonsQueue($guid, $name);
        }
      }
    }
    // Otherwise, add all seasons to the queue.
    else {
      $shows = $this->client->getShows(['fetch-related' => FALSE]);
      foreach ($shows as $show) {
        $count += $this->updateSeasonsQueue($show->id,
          $show->attributes->title);
      }
    }

    if (!$logging['queue']) {
      $this->logger->notice('{count} total seasons have been added to the queue.', ['count' => $count,]);
    }

    return TRUE;
  }

  /**
   * Adds or updates a season node based on API data.
   *
   * @param object $item
   *   An API response object for the season.
   * @param bool $force
   *   Whether to "force" the update. If FALSE (default) and a current
   *   season content node is found with a matching or newer "last_updated"
   *   value, the existing node will not be updated. If TRUE, "last_updated" is
   *   ignored.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Exception
   */
  public function addOrUpdateSeason(object $item, $force = FALSE): void {

    $mappings = $this->config->get('seasons.mappings');
    $node = $this->getOrCreateNode($item->id, $this->getBundleId(), 'seasons');

    $operation = $this->getOperation($item, $node, $mappings, $force);
    $logging = $this->config->get('logging.operations');

    if ($operation == 'configured') {
      $this->logger->error($this->t('Please map the title, id, and updated_at fields for seasons.'));
      return;
    }
    if ($operation == 'skipped') {
      if (!$logging['skipped']) {
        $this->logger->debug($this->t('The season @title has not changed', [
          '@title' => $node->getTitle(),
        ]));
      }
      return;
    }

    // We need to save the node before creating the videos because seasons
    // reference videos and videos reference seasons. Therefore, we also have
    // to set the title of the season. Additionally, we need the guid because
    // addOrUpdateAssetContent() looks up nodes by guid.
    $attributes = $item->attributes;
    $season_number = $attributes->ordinal ?: '';
    $show_title = $attributes->show->attributes->title ?: '';
    if (!empty($attributes->title)) {
      $season_title = $attributes->title;
    }
    elseif (!empty($season_number) && !empty($show_title)) {
      $season_title = $show_title . ": Season " . $season_number;
    }
    else {
      $this->logger->error($this->t('The season ID @id is missing a title, season number, and/or associated show title.', [
        '@id' => $item->id,
      ]));
      return;
    }
    $node->setTitle($season_title);
    if (!empty($item->id) && $mappings['id'] != 'unused') {
      $node->set($mappings['id'], $item->id);
    }
    $node->save();

    // Since we are NOT using the changelog, we must check all the assets for
    // the season to determine if they have been updated.
    if ($video_field = $mappings['video']) {
      if (!empty($video_field) && $video_field !== 'unused' && !empty($attributes->assets)) {
        $videos = $this->client->get('seasons/' . $item->id . "/assets");
        $video_drupal_ids = [];
        foreach ($videos as $video) {
          if ($media_video = $this->assetManager->addOrUpdateAssetContent
          ($video,
            $force)) {
            $video_drupal_ids[] = ['target_id' => $media_video->id()];
          }
        }
        $node->set($video_field, $video_drupal_ids);
      }
    }


    if ($operation == 'ignored') {
      if (!$logging['ignored']) {
        $this->logger->debug($this->t('The season @id was ignored due to a hook_pbs_media_manager_operation_alter() implementation.', [
          '@id' => $item->id,
        ]));
      }
      return;
    }

    if (!empty($item->id) && $mappings['id'] != 'unused') {
      $node->set($mappings['id'], $item->id);
    }
    if (!empty($attributes->ordinal) && $mappings['ordinal'] != 'unused') {
      $node->set($mappings['ordinal'], $attributes->ordinal);
    }
    if (!empty($attributes->description_long) &&
      $mappings['description_long'] != 'unused') {
      $node->set($mappings['description_long'], $attributes->description_long);
    }
    if (!empty($attributes->description_short) &&
      $mappings['description_short'] != 'unused') {
      $node->set($mappings['description_short'], $attributes->description_short);
    }

    $updated_at = self::getLatestUpdatedAt($item);
    $node->set(
      $mappings['updated_at'],
      $updated_at->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT)
    );

    // Add the show reference, if configured.
    $show_ref_field = $this->config->get('seasons.references.show');
    $show_guid = $attributes->show->id;
    $show_bundle = $this->config->get('shows.drupal_show_content');
    if (!empty($show_guid) && !empty($show_bundle) && $show_ref_field != 'unused') {
      $show_node = $this->getNodeByGuid($show_guid, $show_bundle, 'shows');
      if (!empty($show_node) && !empty($show_node->id())) {
        $node->set($show_ref_field, ['target_id' => $show_node->id()]);
      }
    }

    $node->save();

    if (!empty($node->getTitle()) & !empty($operation && !$logging['changed'])) {
      $this->logger->notice($this->t('Season @title has been @operation', [
        '@title' => $node->getTitle(),
        '@operation' => $operation,
      ]));
    }
  }

  /**
   * Query the Seasons endpoint to add items to the queue.
   *
   * @param string $show_guid
   *   A GUID for a specific show.
   * @param string $show_name
   *   The name of the show.
   *
   * @return int
   *   Returns the total number of seasons added per run for logging purposes.
   */
  protected function updateSeasonsQueue(string $show_guid, string $show_name): int {
    $logging = $this->config->get('logging.operations');
    // Initialize counters for tracking items added to queue.
    $count = 0;
    $count_per_show = 0;

    // Get all seasons per show from the API.
    try {
      $seasons = $this->client->getSeasons($show_guid);
    }
    catch (BadRequestException $e) {
      // Only log show errors in ShowManager.
    }

    // Cycle through the response adn add each season to the queue.
    if (isset($seasons) && is_object($seasons)) {
      foreach ($seasons as $season) {
        if ($season_item = $this->client->getSeason($season->id)) {
          $this->getQueue()->createItem($season_item);
          $count++;
          $count_per_show++;
        }
      }
    }
    if ($count_per_show && !$logging['queue']) {
      $this->logger->debug('{count} seasons have been added to the queue for show {show}.', [
        'count' => $count_per_show,
        'show' => $show_name
      ]);
    }

    return $count;

  }

}
