<?php

namespace Drupal\pbs_media_manager;

use DateTime;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\State\StateInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;

class ShowManagerNoChangelog extends ShowManager {

  /**
   * Asset manager.
   *
   * @var \Drupal\pbs_media_manager\AssetManager
   */
  protected AssetManager $assetManager;

  /**
   * Constructor.
   *
   * {@inheritdoc}
   */
  public function __construct(
    ApiClient $client,
    EntityTypeManagerInterface $entity_type_manager,
    ModuleHandlerInterface $module_handler,
    ConfigFactoryInterface $config_factory,
    LoggerChannelInterface $logger,
    QueueFactory $queue_factory,
    StateInterface $state,
    FileSystemInterface $file_system,
    AssetManager $asset_manager
  ) {
    parent::__construct($client, $entity_type_manager, $module_handler,
      $config_factory, $logger, $queue_factory, $state, $file_system);
    $this->assetManager = $asset_manager;
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   */
  public function updateQueue(DateTime $since = NULL): bool {
    $dt_start = new DateTime();
    $logging = $this->config->get('logging.operations');

    // Initialize counter for tracking items added to queue.
    $count = 0;

    // Check to see if we're subscribed to specific shows.
    $subscription = FALSE;
    $show_vocabulary = $this->config->get('shows.show_vocabulary');
    if (!empty($show_vocabulary) && $show_vocabulary != 'unused') {
      $subscription = TRUE;
      // Get the shows.
      $show_guids = $this->getSubscribedShows($show_vocabulary);
    }

    // If only subscribed to certain shows, only add those shows to the queue.
    if ($subscription && !empty($show_guids)) {
      foreach ($show_guids as $guid => $name) {
        $show_item = $this->client->getShow($guid);
        $this->getQueue()->createItem($show_item);
        $count++;
      }
    }
    // Otherwise, add all shows to the queue.
    else {
      $shows = $this->client->getShows(['fetch-related' => FALSE]);
      foreach ($shows as $show) {
        $this->getQueue()->createItem($show);
        $count++;
      }
    }

    // Update the logger with the total number of shows added.
    if (!$logging['queue']) {
      $this->logger->notice('{count} total shows have been added to the queue.', ['count' => $count ]);
    }

    $this->setLastUpdateTime($dt_start);

    return TRUE;
  }

  /**
   * Adds or updates a Show node based on API data.
   *
   * @param object $item
   *   An API response object for the Show.
   * @param bool $force
   *   Whether to "force" the update. If FALSE (default) and a current
   *   show content node is found with a matching or newer "last_updated"
   *   value, the existing node will not be updated. If TRUE
   *   "last_updated" is ignored.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Exception
   */
  public function addOrUpdateShow(object $item, $force = FALSE): void {

    $mappings = $this->config->get('shows.mappings');
    $node = $this->getOrCreateNode($item->id, $this->getBundleId(), 'shows');
    $operation = $this->getOperation($item, $node, $mappings, $force);
    $logging = $this->config->get('logging.operations');

    // Do not proceed if mappings are missing.
    if ($operation == 'configured') {
      $this->logger->error($this->t('Please map the title, id, and updated_at fields for shows.'));
      return;
    }
    // Do not consider videos for ignored shows.
    if ($operation == 'ignored') {
      if (!$logging['ignored']) {
        $this->logger->debug($this->t('The show @id was ignored due to a hook_pbs_media_manager_operation_alter() implementation.', [
          '@id' => $item->id,
        ]));
      }

      return;
    }

    // We need to save the node before creating the videos because shows
    // reference videos and videos reference shows. Therefore, we also have
    // to set the title of the show. Additionally, we need the guid because
    // addOrUpdateAssetContent() looks up nodes by guid.
    $attributes = $item->attributes;
    if (empty($attributes->title)) {
      return;
    }
    $node->setTitle($attributes->title);
    if (!empty($item->id) && $mappings['id'] != 'unused') {
      $node->set($mappings['id'], $item->id);
    }
    $node->save();

    // Since we are NOT using the changelog, we must check all the assets for
    // the show to determine if they have been updated.f
    if ($video_field = $mappings['video']) {
      if (!empty($video_field) && $video_field !== 'unused' && !empty($attributes->assets)) {
        $videos = [];
        try {
          $videos = $this->client->get('shows/' . $item->id . "/assets");
        }
        catch (Exception $e) {
          $this->logger->error(t('The show ID @id asset request failed. Msg: @msg', [
            '@id' => $item->id,
            '@msg' => $e->getMessage(),
          ]));
        }
        $video_drupal_ids = [];
        foreach ($videos as $video) {
          if ($media_video = $this->assetManager->addOrUpdateAssetContent
          ($video,
            $force)) {
            $video_drupal_ids[] = ['target_id' => $media_video->id()];
          }
        }
        $node->set($video_field, $video_drupal_ids);
      }
    }

    if ($operation == 'skipped') {
      if (!$logging['skipped']) {
        $this->logger->debug($this->t('The show @title has not changed', [
          '@title' => $node->getTitle(),
        ]));
      }
      return;
    }

    // Get the call sign field and value.
    $call_sign_field = $mappings['call_sign'];
    $stations = array_column($attributes->audience, 'station');
    $call_sign = '';
    foreach ($stations as $station) {
      if (isset($station->type) && $station->type == 'station') {
        $call_sign = $station->attributes->call_sign;
      }
    }

    if (!empty($call_sign_field) && $mappings['call_sign'] != 'unused') {
      $node->set($call_sign_field, $call_sign);
    }
    if (!empty($item->id) && $mappings['id'] != 'unused') {
      $node->set($mappings['id'], $item->id);
    }
    if (!empty($attributes->tms_id) && $mappings['tms_id'] != 'unused') {
      $node->set($mappings['tms_id'], $attributes->tms_id);
    }
    if (!empty($attributes->description_long) &&
      $mappings['description_long'] != 'unused') {
      $node->set($mappings['description_long'], $attributes->description_long);
    }
    if (!empty($attributes->description_short) &&
      $mappings['description_short'] != 'unused') {
      $node->set($mappings['description_short'], $attributes->description_short);
    }
    if (!empty($attributes->nola) &&
      $mappings['nola'] != 'unused') {
      $node->set($mappings['nola'], $attributes->nola);
    }

    if (!empty($attributes->drm_enabled) && $mappings['drm_enabled'] != 'unused') {
      $node->set($mappings['drm_enabled'], $attributes->drm_enabled);
    }

    $updated_at = self::getLatestUpdatedAt($item);
    $node->set(
      $mappings['updated_at'],
      $updated_at->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT)
    );

    // Add the genre, if mapped.
    if (!empty($attributes->genre) && $genre_field = $mappings['genre']) {
      if ($genre_field != 'unused') {
        $node->set($genre_field, $this->getOrAddGenre(
          $attributes->genre->id,
          $attributes->genre->title,
          $attributes->genre->slug
        ));
      }
    }

    // Logic to handle the links if the Social Media Links Block and Field
    // module (https://www.drupal.org/project/social_media_links) is enabled.
    // TODO: Add additional options to handle the links array.
    $links_field = $mappings['links'];
    if (!empty($attributes->links) &&
      $links_field != 'unused' &&
      $this->moduleHandler->moduleExists('social_media_links_field')) {
      foreach ($attributes->links as $link) {
        if (isset($node->{$links_field}->platform_values)) {
          $link_value = $link->value;
          // All profiles except newsletters & email have a hard coded base URL.
          if (!in_array($link->profile, ['newsletter', 'email'])) {
            $link_value = parse_url($link->value, PHP_URL_PATH);
            // Remove beginning and trailing slashed.
            $link_value = preg_replace('/^\//', '', $link_value);
            $link_value = preg_replace('/\/$/', '', $link_value);
          }
          $platform_values[$link->profile]['value'] = $link_value;
        }
      }
      $node->set($links_field, ['platform_values' => $platform_values]);
    }

    // Process images.
    if ($images = $attributes->images) {
      foreach ($images as $image) {
        if (!empty($mappings[$image->profile]) && $image_field = $mappings[$image->profile]) {
          if ($image_field != 'unused') {
            // Check to see if this item has an image already.
            $media_image = NULL;
            if ($node->get($image_field)->target_id) {
              // Get the existing image entity.
              $image_id = $node->get($image_field)->target_id;
              $media_storage = $this->entityTypeManager->getStorage('media');
              $media_image = $media_storage->load($image_id);
            }
            if ($media_id = $this->addOrUpdateMediaImage(
              $image,
              $attributes->title . ": " . $image->profile,
              'show',
              $media_image
            )) {
              $node->set($image_field, ['target_id' => $media_id]);
            }
          }
        }
      }
    }

    $node->save();

    if (!empty($node->getTitle()) & !empty($operation && !$logging['changed'])) {
      $this->logger->notice($this->t('Show @title has been @operation', [
        '@title' => $node->getTitle(),
        '@operation' => $operation,
      ]));
    }
  }


}
